import 'dart:convert';

import 'package:http/http.dart' as http;
import 'package:uuid/uuid.dart';

import 'odoo_version.dart';

abstract class OdooConnector {
  http.Client _client;
  String _serverURL;
  Map<String, String> _headers = {};
  bool _debugRPC = false;
  Map<String, Function> _callbackFunctions = {};
  OdooVersion odooVersion = new OdooVersion();
  List _databases = [];
  String _sessionId;

  OdooConnector(String serverURL) {
    this._serverURL = serverURL;
    _client = http.Client();
  }

  void setSessionId(String sessionId) {
    _sessionId = sessionId;
  }

  void connect([Function(bool isConnected, OdooVersion version) callback]) {
    getVersionInfo((result) {
      odooVersion = result;
      print("Connected to ${_serverURL} : ${odooVersion.toString()}");
      getDatabases((databases) {
        if (callback != null) {
          callback(true, odooVersion);
        }
      });
    });
  }

  getVersionInfo([Function(OdooVersion) callback]) {
    if (callback != null) {
      var url = createPath("/web/webclient/version_info");
      newRequest(url, createPayload({}), (result) {
        odooVersion = OdooVersion().parse(result);
        callback(odooVersion);
      });
    } else {
      return odooVersion;
    }
  }

  // Get database lists
  getDatabases([Function(List) callback]) {
    if (_databases.length > 0 && callback == null) {
      return _databases;
    }
    OdooVersion version = getVersionInfo();
    if (version.getMajorVersion() != null) {
      String url = getServerURL();
      var params = {};
      if (version.getMajorVersion() == 9) {
        url = createPath("/jsonrpc");
        params["method"] = "list";
        params["service"] = "db";
        params["args"] = [];
      } else if (version.getMajorVersion() >= 10) {
        url = createPath("/web/database/list");
        params["context"] = {};
      } else {
        url = createPath("/web/database/get_list");
        params["context"] = {};
      }
      newRequest(url, params, (result) {
        _databases = result.getResult();
        if (callback != null) {
          callback(_databases);
        }
      });
    } else {
      connect((flag, version) {
        if (callback != null) {
          callback(getDatabases());
        }
      });
    }
  }

  void debugRPC(bool debug) {
    _debugRPC = debug;
  }

  String getServerURL() {
    return _serverURL;
  }

  String createPath(String path) {
    return _serverURL + path;
  }

  Map createPayload(Map params) {
    return {
      "jsonrpc": "2.0",
      "method": "call",
      "params": params,
      "id": new Uuid().v1()
    };
  }

  void newRequest(String url, Map payload, Function(OdooResponse) callback) {
    var body = json.encode(payload);
    _callbackFunctions[payload["id"]] = callback;
    _headers["Content-type"] = "application/json";
    if (_sessionId != null) {
      _headers['Cookie'] = "session_id=" + _sessionId;
    }
    if (_debugRPC) {
      print("-------------------------------------------");
      print("REQUEST: ${url}");
      print("PAYLOD : ${payload}");
      print("HEADERS: ${_headers}");
      print("-------------------------------------------");
    }
    _client.post(url, body: body, headers: _headers).then((response) {
      _updateCookies(response);
      if (_debugRPC) {
        print("============================================");
        print("STATUS_C: ${response.statusCode}");
        print("RESPONSE: ${response.body}");
        print("============================================");
      }
      Map data = json.decode(response.body);
      Function responseCallback = _callbackFunctions[data['id']];
      if (responseCallback != null) {
        var responseObj = new OdooResponse(data, response.statusCode);
        responseCallback(responseObj);
        _callbackFunctions.remove(data['id']);
      }
    }).catchError((error) {
      print("ERROR: ${error}");
    });
  }

  _updateCookies(http.Response response) {
    String rawCookie = response.headers['set-cookie'];
    if (rawCookie != null) {
      int index = rawCookie.indexOf(';');
      _headers['Cookie'] =
          (index == -1) ? rawCookie : rawCookie.substring(0, index);
    }
  }
}

class OdooResponse {
  var _result, _statusCode;

  OdooResponse(Map result, int statusCode) {
    _result = result;
    _statusCode = statusCode;
  }

  bool hasError() {
    return _result['error'] != null;
  }

  Map getError() {
    return _result['error'];
  }

  String getErrorMessage() {
    if (hasError()) {
      return _result['error']['message'];
    }
    return null;
  }

  int getStatusCode() {
    return _statusCode;
  }

  dynamic getResult() {
    return _result['result'];
  }
}
